/* File:	  Compiler.P
** Author(s): Ninghui Li
** Contact:   ninghui@cs.nyu.edu
** 
** Copyright (C)  2000, Ninghui Li
**
** XD1LP is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XD1LP is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XD1LP; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/************************************************************************
                      File Compiler.P

  This file contains the compiler that compiles rules form D1LP internal 
  formats to OLP internal formats.  

  This compilation process adds the length parameter, substitutes 
  principal structures with dummy principals, and generate new rules for 
  delegation propogation.

  The main predicate is compile_rule/2.
*************************************************************************/

:- export compile_rule/4, compile_query/3. 

:- import append/3, member/2 from basics.

/* Internal format
  dl_holds(Issuer, BAtom)
  dl_delegates(Issuer, BAtom, Dep, Dele)
 */

 %% compile_rule(+D1LPRule, -OLPRuleSet)
 %%  compiles the rule D1LPRule to an OLPRuleSet. 
    %% This clause handles d1lp rules whose heads are holds statements
compile_rule((dl_holds(Issuer, BAtom) :- Body), 
               [srule(holds(Issuer, BAtom, 1), NewBody) 
                | NewRules], 0, IL) :-
    compile_body(Body, NewBody, NewRules, BIL),
    (var(Issuer) -> IL=BIL ; IL=[Issuer | BIL]).
    %% This clause handles d1lp rules whose heads are delegates statements
compile_rule((dl_delegates(Issuer, BAtom, Dep, Dele) :- Body),
               [srule(NewHead, NewBody), 
                srule(delegates(Issuer, BAtom, Dep, VarDele2, 1),
                     (is_prin_or_var(VarDele2),
                      NewHead, 
                      delegates(Dele, BAtom, '*', VarDele2, 0))),
                rule(holds(Issuer, BAtom, Len1), 
                     (NewHead, 
                      special_minus_one_min(Len1, Dep, VarLen1), 
                      holds(Dele, BAtom, VarLen1))),
                rule(delegates(Issuer, BAtom, VarDep3, VarDele3, Len3),
                     (is_prin_or_var(VarDele3),
                      special_less_tran(VarDep3, Dep),
                      NewHead, 
                      special_minus_one(Len3, VarLen3), 
                      delegates(Dele, BAtom, VarDep3, VarDele3, VarLen3)))
                | NewRules], Dep, IL) :-
    compile_body(Body, NewBody, NewRules, BIL),    %% compiles body
    compile_head(dl_delegates(Issuer, BAtom, Dep, Dele), NewHead),
    (var(Issuer) -> IL=BIL ; IL=[Issuer | BIL]).
compile_rule((dl_speaksfor(Dele, BAtom, Issuer) :- Body),
               [srule(delegates(Issuer, BAtom, '*', Dele, 0), NewBody),
                rule(holds(Issuer, BAtom, Len),
                     (delegates(Issuer, BAtom, '*', Dele, 0),
                      holds(Dele, BAtom, '*'))),
                rule(delegates(Issuer, BAtom, VarDep, VarDele, Len),
                     (is_prin_or_var(VarDele),
                      delegates(Issuer, BAtom, '*', Dele, 0),
                      delegates(Dele, BAtom, VarDep, VarDele, Len)))
                | NewRules], 0, IL) :-
    compile_body(Body, NewBody, NewRules, BIL),
    (var(Issuer) -> IL=BIL ; IL=[Issuer | BIL]).

 %% compile_head(+D1LPHead, -OLPHead)
 %%  returns a new head to subsitute the old one; the new head is  
 %%  changes a complex delegatee to a dummy principal.
compile_head(dl_delegates(Issuer, BAtom, Dep, OldDele),
               delegates(Issuer, BAtom, Dep, NewDele, 1)) :-
    get_principal(OldDele, NewDele).


compile_query(OldQuery, NewQuery, NewRules) :-
    compile_body(OldQuery, NewQuery, NewRules, _NIL).

 %% compile_body(+OldBody, -NewBody, -NewRules, -NewPrincipals)
 %%  compiles the body of one rule, mainly adding '*' as Len, 
 %%  also deals with delegation to conjunction of principals. 
compile_body(true, true, [], []).
compile_body((A, B), (NA, NB), NewRules, NewIL) :- 
    compile_body(A, NA, NewRulesA, NewILA), 
    compile_body(B, NB, NewRulesB, NewILB),
    append(NewRulesA, NewRulesB, NewRules),
    append(NewILA, NewILB, NewIL).

compile_body((A; B), (NA; NB), NewRules, NewIL) :- 
    compile_body(A, NA, NewRulesA, NewILA),
    compile_body(B, NB, NewRulesB, NewILB),
    append(NewRulesA, NewRulesB, NewRules),
    append(NewILA, NewILB, NewIL).

compile_body(dl_holds(Issuer, BAtom), holds(Issuer, BAtom, '*'), [], []).
compile_body(dl_delegates(Issuer, BAtom, Dep, Dele),
             delegates(Issuer, BAtom, Dep, Dele, '*'), [], []) :-
    is_prinorvar(Dele). 
compile_body(dl_delegates(Issuer, BAtom, Dep, (PrinS1,PrinS2)),
             delegates(Issuer, BAtom, Dep, NewDele, '*'),
             NewRules, [NewDele]) :-
    new_principal(NewDele),
    add_delegates((PrinS1,PrinS2), dl_delegates(BAtom, '*', NewDele, 0), NewRules).

 %% add_delegates(+PrinConjuncts, +D1LPDelegates, -AdditionalRules)
 %%  is used to create the delegation rules from all principals in a 
 %%  conjunct to a dummy principal
add_delegates((PrinSA,PrinSB), dl_delegates(BAtom, Dep, Dele, Len), NewRules) :-
    add_delegates(PrinSA, dl_delegates(BAtom, Dep, Dele, Len), NewRulesA),
    add_delegates(PrinSB, dl_delegates(BAtom, Dep, Dele, Len), NewRulesB),
    append(NewRulesA, NewRulesB, NewRules).
add_delegates(Prin, dl_delegates(BAtom, Dep, Dele, Len), 
              [srule(delegates(Prin, BAtom, Dep, Dele, Len), true)]) :-
    is_prinorvar(Prin).

 %% get_principal(+OldDele, -NewDele) 
 %%  returns the old delegatee if it is a principal, 
 %%  returns a new principal otherwise
get_principal(Prin, Prin) :- is_prinorvar(Prin), !.
get_principal(_ComplexDele, Prin) :- new_principal(Prin).

is_prinorvar(Prin) :- is_principal(Prin); var(Prin).

is_principal(Prin) :- atom(Prin).

new_principal(Prin) :- new_atom(Prin).

new_atom(Atom) :-
    retract(new_atom_counter(N)),
    NewN is N + 1, 
    assert(new_atom_counter(NewN)),
    number_codes(NewN, NString),
    append("tmp_atom_", NString, AtomString),
    atom_codes(Atom, AtomString).
