/* File:	  Tokenizer.P
** Author(s): Ninghui Li
** Contact:   ninghui@cs.nyu.edu
** 
** Copyright (C)  2000, Ninghui Li
**
** XD1LP is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XD1LP is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XD1LP; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
%% The following is a lexical scanner.  It can convert a string to a list 
%% of tokens.  It will strip out comments, that start with a %, //, 
%% Tokens include atoms, numbers, strings, variables. 
%%   Variables must start with a question mark.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- export string_to_tokens/2, token_to_string/2.

 %% string_to_tokens(+String, -Tokens)
 %% converts a string to a list of tokens
string_to_tokens(Str, Tokens) :- my_string_to_tokens(Str, Tokens).

my_string_to_tokens([], []) :-  !.
my_string_to_tokens(Sin, RTs) :-
    one_token(Sin, T, Stmp), 
    ((T = end_of_string) -> 
        (RTs = []);
        (my_string_to_tokens(Stmp, Ts), RTs=[T|Ts])).

 %% one_token(+String, -Token, -RemainingString)
 %%   get one token from the input, string returns the remaning string
one_token([], end_of_string, []) :- !.
one_token([-1], end_of_string, []) :- !.  %% eof 

/* Don't allow this kind of comments for now
one_token([35|Sin], T, Sout) :-		    %% comment starting with a '#' 
    !, consume_rest_of_line(Sin, Stmp),
    one_token(Stmp, T, Sout).
*/

one_token([37|Sin], T, Sout) :-		    %% comment starting with a '%' 
    !, consume_rest_of_line(Sin, Stmp),
    one_token(Stmp, T, Sout).

one_token([47|Sin], T, Sout) :-		    %% stuff starting with a '/', i.e, 
    !, after_slash(Sin, T, Sout).       %% //, /* */, or just a /

/* Don't know why we don't deal with string now
one_token([34|Sin], string(T), S) :-    %% sees '"', get a string
    !, rest_of_string(Sin, T, S).
*/

one_token([63|Sin], var(T), Sout) :-    %% sees '?', start a variable 
    !, rest_of_token(Sin, Cs, Sout),
    atom_codes(T, Cs).                  %% use atoms for variables

one_token([C|S], T, S) :-		        %% other single characters 
    single_character(C, T), !.          %% old code:  name(T, [C]). 
    
one_token([C|Sin], T, Sout) :-		    %% This handles an atom or an 
    in_token(C), !,                     %% integer
    rest_of_token(Sin, Cs, Sout),
    token_codes(T, [C|Cs]).

one_token([_|Sin], T, Sout) :- 		    %% can't undetsand current char, 
    one_token(Sin, T, Sout).            %% ignore it and move on

 
 %% rest_of_token(+String, -TokenString, -RemainingString)
rest_of_token([C|Sin], [C|Cs], Sout) :-
    in_token(C), !,                  %% keep reading if the current char 
    rest_of_token(Sin, Cs, Sout).    %% is a legal char in a token
rest_of_token(S, [], S).             %% otherwise, stop


 %% after_slash(+String, -Token, -RemainingString)
after_slash([47|Sin], T, Sout) :-       %% if '/' is followed by '/', 
    !, consume_rest_of_line(Sin, Stmp), %% consume the rest of the line 
    one_token(Stmp, T, Sout).           %% and read another token

/* Do not allow C style comments yet, since they make it impossible to 
   tokenize one line at a time. 
after_slash([42|Sin], T, Sout) :-		    %% if '/' is followed by '*', 
    !, consume_rest_of_comment(Sin, Stmp),  %% consume the rest of the 
    one_token(Stmp, T, Sout).               %% comment
*/
after_slash(S, T, S) :-		            %% Otherwise, return symbol for '/'
    name(T, [47]).


 %% consume_rest_of_line(+String, -RemainingString)
 %%  throws out the rest of the current line from String, 
 %%  returns the remaining string. 
consume_rest_of_line([13|S], S) :-      %% reach end of line, stop
    !.
consume_rest_of_line([_C|R1], R) :-     %% otherwise, keep consuming
    consume_rest_of_line(R1, R).

/* Now don't allow C-style comments
 %% consume_rest_of_comment(+String, -RemainingString)
 %%  throws out the rest of the comment, which ends with * and /
 %%  returns the remaining string. 
consume_rest_of_comment([42|Sin], Sout) :-
    !, try_end_comment(Sin, Sout).      %% sees a '*' in comment, try end
consume_rest_of_comment([_|Sin], Sout) :-
    consume_rest_of_comment(Sin, Sout). %% Otherwise, keep consuming
    
 %% try_end_comment(+String, -RemainingString)
 %%  tries to see a to end the comment
 %%  returns the remaining string. 
try_end_comment([47|S], S) :-  !.       %% sees a '/', end
try_end_comment([42|Sin], Sout) :-      %% sees a '*', try again
    !, try_end_comment(Sin, Sout).
try_end_comment([_|Sin], Sout) :-       %% otherwise, go back to consuming
    consume_rest_of_comment(Sin, Sout).
 */


%% These characters form words on their own 

single_character(58, ':').
single_character(46, '.').
single_character(40, '(').
single_character(41, ')').
single_character(42, '*').
single_character(43, '-').
single_character(44, ',').
single_character(45, '-').
single_character(59, ';').
single_character(60, '<').
single_character(61, '=').
single_character(62, '>').
single_character(92, '\').
single_character(91, '[').
single_character(93, ']').
single_character(123, '{').
single_character(125, '}').
single_character(94, '^').

%% single_character(33, '!').
%% single_character(63, '?').
%% single_character(126, '~').
%% single_character(47, '/').


%% These characters can appear within a word 

in_token(39).		        /* ' */
in_token(95).		        /* _ */
in_token(C) :- C>96, C<123.	/* a b ... z */
in_token(C) :- C>64, C<91.	/* A B ... Z */
in_token(C) :- C>47, C<58.	/* 0 1 ... 9 */

token_codes(T, CharList) :- name(T, CharList).

/* don't know why it is here
 %% token_from_string(-Token, +String)
 %%   get a token from its string representation
token_from_string(Token, CharList) :- 
    is_charlist(CharList),
    var(Token), 
    ( number_codes(Token, CharList) -> true ; 
      atom_codes(Token, CharList) ).
 */

 %% token_to_string(+Token, -String)
 %%  get a token's string representation
token_to_string(var(VarName), [63 | CharList]) :- 
    !, token_to_string(VarName, CharList).
token_to_string(Token, CharList) :- 
    atomic(Token),
    ( number(Token) -> number_codes(Token, CharList) ; 
      atom_codes(Token, CharList) ).
